"""
Relational database can be used as a cache or a persistence storage.
Actually, implement a cache is enough.
"""

from .encoder import ValueEncoder, NoModifyEncoder
from .decoder import ValueDecoder, NoModifyDecoder
from .adapter import ValueAdapter, _BUILTIN_ADAPTER_BUILDERS
from .config import Operation, BaseTableConfig, RelationalDbScopeConfig, SingleTableConfig
from .sql import (
    SqlBuilder,
    SimpleSqlBuilder,
    QmarkSqlBuilder,
    NumericSqlBuilder,
    NamedSqlBuilder,
    FormatSqlBuilder,
    PyformatSqlBuilder,
    DEFAULT_SQL_BUILDERS,
)
from .cache import BaseRelationalDbCache, RelationalDbCache, SingleTableCache



from abc import ABC, abstractmethod
from typing import Any, Dict, Sequence, Tuple
import collections
from .config import _EXECUTE_PARAM_TYPE, Operation, BaseTableConfig

# sql syntax may diff for different db, each backend should implement it's sql_builder
class SqlBuilder(ABC):
    """
    Build select, update, delete sql for relational database.
    It's used by relational cache to manipulate data.
    """

    def build_select_all_table_operation(self) -> Operation:
        """
        Generate sql to select all tables in the database.
        Table name as first selected column.
        """
        raise NotImplementedError()

    def build_select_table_columns_operation(self, table: str, filter_uniq=False) -> Operation:
        """
        Generate sql to select all columns for given table.
        Column name as first selected result.
        :param table:           the queried table
        :param filter_uniq:     return only id column
        """
        raise NotImplementedError()

    def get_column_name_from_result(self, result: Sequence[Any]) -> str:
        """
        Get column name from result of operation generated by `build_select_table_columns_operation()`.
        """
        # assume first result is column name.
        return result[0]

    @abstractmethod
    def build_create_table_operation(self, *configs: BaseTableConfig, check_exists=True):
        """
        Generate sql to create a table.
        """

    @abstractmethod
    def build_select_operation(self, config: BaseTableConfig, key: Sequence[Any] = None, limit: int = -1,
                               select_columns: Sequence[str] = None, condition_columns: Sequence[str] = None,
                               distinct=False) -> Operation:
        """
        Generate sql to select row for given key.
        :param config:          scope configuration
        :param key:             uniq key for the row, maybe `None` to select all rows in the table
        :param limit:           limit number
        :param select_columns:  columns to be selected; if not set, select all in config
        :param condition_columns:  columns for where condition; if not set, use uniq columns
        :param distinct:        whether distinct select columns
        """

    @abstractmethod
    def build_update_operation(self, config: BaseTableConfig, key: Sequence[Any], value: Dict[str, Any]) -> Operation:
        """
        Generate sql to update or insert row for given key with given value.
        :param config:      scope configuration
        :param key:         uniq key for the row
        :param value:       column values for given uniq key
        """

    @abstractmethod
    def build_delete_operation(self, config: BaseTableConfig, key: Sequence[Any] = None) -> Operation:
        """
        Generate sql to delete row for given key.
        :param config:      scope configuration
        :param key:         uniq key for the row, maybe `None` to delete all rows in the table, be careful
        """

class SimpleSqlBuilder(SqlBuilder, ABC):
    """
    A simple implement of SqlBuilder.
    """

    # whether to pass parameters as a list
    list_parameter = False

    # if column type not specified in the config, use this value as default
    # should be overwritten for different backends
    default_column_type = 'text'

    def _get_column_type(self, col: str, config: BaseTableConfig) -> str:
        return config.get_column_type(col) or self.default_column_type

    def _gen_column_def(self, col: str, config: BaseTableConfig) -> str:
        t = self._get_column_type(col, config)
        s = f"{col} {t}"
        # if col in config.uniq_columns:
        # s += " UNIQUE"
        # s += ' PRIMARY KEY'
        return s

    def _gen_column_defines(self, config: BaseTableConfig):
        columns = [self._gen_column_def(x, config) for x in config.columns_with_id]
        columns += [f" PRIMARY KEY ({','.join(config.uniq_columns)})"]
        return ','.join(columns)

    def _gen_create_table_sql(self, config: BaseTableConfig, check_exists=True) -> str:
        sql = "CREATE TABLE"
        if check_exists:
            sql += " IF NOT EXISTS"
        sql += f" {config.table} ({self._gen_column_defines(config)})"
        return sql

    def build_create_table_operation(self, *configs: BaseTableConfig, check_exists=True) -> Operation:
        lines = [self._gen_create_table_sql(x) + ";" for x in configs]
        return Operation(statement="\n".join(lines), many=len(configs) > 1)

    @abstractmethod
    def get_variable_placeholder(self, name: str = None, order: int = None, value: Any = None) -> str:
        """
        Get variable placeholder in operation template.
        :param name:                variable name, used for named and pyformat style
        :param order:               order in the parameters, used for numeric style, start from 1
        :param value:               variable value, for ansi c printf format codes without '%', such as 's' or 'f'
                                    used for format and pyformat style
        """

    def config_variable(self, name: str = None, order: int = None, value: Any = None,
                        variable_mapping: Dict[str, Any] = None) -> str:
        """
        Configure a variable with name, order, value.
        :param name:                variable name, used for named and pyformat style
        :param order:               order in the parameters, used for numeric style, start from 1
        :param value:               variable value, for ansi c printf format codes without '%', such as 's' or 'f'
                                    used for format and pyformat style
        :param variable_mapping:    name -> value, used to save value to variable in this method
        :return:                    a placeholder in operation template
        """
        if variable_mapping is not None:
            variable_mapping[name] = value
        return self.get_variable_placeholder(name=name, order=order, value=value)

    def normalize_variable_values(self, variable_values: Dict[str, Any],
                                  variable_names: Sequence[str] = None) -> _EXECUTE_PARAM_TYPE:
        """
        All parameters are saved in a dict first, in this method, it can be changed to a list.
        Element in variable_names should be included in variable_values.
        """
        if self.list_parameter:
            if variable_names is not None:
                return [variable_values[x] for x in variable_names]
            return list(variable_values.values())
        return variable_values

    def build_select_operation(self, config: BaseTableConfig, key: Sequence[Any] = None, limit: int = -1,
                               select_columns: Sequence[str] = None, condition_columns: Sequence[str] = None,
                               distinct=False) -> Operation:
        if select_columns is None:
            # only select configured columns; if id not configured, ignore it
            select_columns = config.columns
        elif isinstance(select_columns, str):
            select_columns = [select_columns]
        stmt = "SELECT"
        if distinct:
            stmt += f" DISTINCT"
        stmt += f" {','.join(select_columns)} FROM {config.table}"
        variable_values = collections.OrderedDict()
        if key is not None:
            if condition_columns is None:
                # use uniq columns as where condition
                condition_columns = config.uniq_columns
            elif isinstance(condition_columns, str):
                condition_columns = [condition_columns]
            assert len(condition_columns) == len(key)
            condition = []
            for i, (name, k) in enumerate(zip(condition_columns, key), 1):
                placeholder = self.config_variable(name=name, order=i, value=k, variable_mapping=variable_values)
                condition.append(f"{name} = {placeholder}")
            condition = ' AND '.join(condition)
            stmt += f" WHERE {condition}"
        if limit > 0:
            stmt += f" LIMIT {limit}"
        return Operation(statement=stmt, parameters=self.normalize_variable_values(variable_values))

    def _gen_update_statement(self, config: BaseTableConfig, value: Dict[str, Any]) -> Tuple[str, bool]:
        """
        Generate statement for Operation, data of uniq column have added to value.
        :return:    (statement, many or not)
        """
        # sql is writen based of syntax of sqlite, maybe is incapable with other database
        first = ','.join(self.config_variable(name=k, order=i, value=v)
                         for i, (k, v) in enumerate(value.items(), 1))
        second = ','.join(f'{k}={self.config_variable(name=k, order=i, value=v)}'
                          for i, (k, v) in enumerate(value.items(), 1))
        return (f"INSERT INTO {config.table}({','.join(value.keys())})"
                f" VALUES ({first})"
                f" ON CONFLICT({','.join(config.uniq_columns)}) DO UPDATE"
                f" SET {second}"), False

    def _gen_update_variables(self, value: Dict[str, Any]) -> Tuple[Dict[str, Any], Sequence[str]]:
        """
        Generate variable values and name list used when execute().
        :param value:   value with uniq column
        :return:    (variable_values, variable_name_list)
        """
        return value, list(value.keys()) * 2

    def build_update_operation(self, config: BaseTableConfig, key: Sequence[Any], value: Dict[str, Any]) -> Operation:
        # add uniq column to value
        assert len(config.uniq_columns) == len(key)
        value = dict(value)
        for name, k in zip(config.uniq_columns, key):
            value[name] = k

        # gen statement and parameters
        stmt, many = self._gen_update_statement(config=config, value=value)
        variable_values, variable_names = self._gen_update_variables(value)
        parameters = self.normalize_variable_values(variable_values, variable_names)

        return Operation(statement=stmt, parameters=parameters, many=many)

    def build_delete_operation(self, config: BaseTableConfig, key: Sequence[Any] = None) -> Operation:
        stmt = f"DELETE FROM {config.table}"
        variable_values = collections.OrderedDict()
        if key is not None:
            assert len(config.uniq_columns) == len(key)
            condition = []
            for i, (name, k) in enumerate(zip(config.uniq_columns, key), 1):
                placeholder = self.config_variable(name=name, order=1, value=k,
                                                   variable_mapping=variable_values)
                condition.append(f"{name} = {placeholder}")
            condition = ' AND '.join(condition)
            stmt += f" WHERE {condition}"
        return Operation(statement=stmt, parameters=self.normalize_variable_values(variable_values))

class QmarkSqlBuilder(SimpleSqlBuilder):

    list_parameter = True

    def get_variable_placeholder(self, name: str = None, order: int = None, value: Any = None) -> str:
        return "?"

class NumericSqlBuilder(SimpleSqlBuilder):

    list_parameter = True

    def get_variable_placeholder(self, name: str = None, order: int = None, value: Any = None) -> str:
        return f":{order}"

class NamedSqlBuilder(SimpleSqlBuilder):

    list_parameter = False

    def get_variable_placeholder(self, name: str = None, order: int = None, value: Any = None) -> str:
        return f":{name}"

class FormatSqlBuilder(SimpleSqlBuilder):

    list_parameter = True

    def get_value_type(self, value: Any) -> str:
        if isinstance(value, str):
            return 's'
        if isinstance(value, int):
            return 'd'
        if isinstance(value, float):
            return 'f'
        return 's'

    def get_variable_placeholder(self, name: str = None, order: int = None, value: Any = None) -> str:
        return f"%{self.get_value_type(value)}"

class PyformatSqlBuilder(FormatSqlBuilder):

    list_parameter = False

    def get_variable_placeholder(self, name: str = None, order: int = None, value: Any = None) -> str:
        return f"%({name}){self.get_value_type(value)}"

# paramstyle -> builder
DEFAULT_SQL_BUILDERS = {
    "qmark": QmarkSqlBuilder(),
    "numeric": NumericSqlBuilder(),
    "named": NamedSqlBuilder(),
    "format": FormatSqlBuilder(),
    "pyformat": PyformatSqlBuilder(),
}




# -*- coding: utf-8 -*-

from .relational import (
    SqlBuilder,
    SimpleSqlBuilder,
    QmarkSqlBuilder,
    NumericSqlBuilder,
    NamedSqlBuilder,
    FormatSqlBuilder,
    PyformatSqlBuilder,
    RelationalDbScopeConfig,
    RelationalDbCache,
)
from .sqlite import (
    SqliteSqlBuilder,
    SqliteCache,
)
try:
    from .postgresql import (
        PostgresqlBuilder,
        PostgresqlScopeConfig,
        PostgresqlCache,
    )
except ImportError:
    # psycopg maybe not installed
    pass
try:
    from .mysql import (
        MysqlSqlBuilder,
        MysqlScopeConfig,
        MysqlCache,
    )
except ImportError:
    # pymysql or mysql.connector should be installed
    pass




from typing import (
    Sequence, Dict, Union, Literal, Callable,
    Any, Optional, Hashable,
)
import dataclasses
import warnings
from .encoder import ValueEncoder, NoModifyEncoder
from .decoder import ValueDecoder, NoModifyDecoder
from .adapter import ValueAdapter, _BUILTIN_ADAPTER_BUILDERS

_EXECUTE_PARAM_TYPE = Union[Sequence[Any], Dict[str, Any]]

@dataclasses.dataclass
class Operation:
    """
    define an operation for execute() or executemany()
    """
    statement: str       # arg operation for cursor.execute()
    parameters: _EXECUTE_PARAM_TYPE = dataclasses.field(default_factory=list)  # arg parameters for cursor.execute()

    # many statements in template or not, if True, use cursor.executemany() instead of cursor.execute()
    many: bool = False

class BaseTableConfig:
    def __init__(
        self,
        table: str,
        uniq_columns: Sequence[str] = ('id',),
        columns: Sequence[str] = ('data',),
        column_types: Dict[str, str] = None,
        value_adapter: Union[Literal['default', 'immutable', 'single', 'tuple', 'list', 'json'], ValueAdapter] = 'default',
        default_column_type: str = None,
        encoder: Union[None, ValueEncoder, Callable[[Any], Dict[str, Any]]] = None,
        decoder: Union[None, ValueDecoder, Callable[[Dict[str, Any]], Any]] = None,
    ):
        """
        :param table:
        :param uniq_columns:    unique columns to identify rows
        :param columns:         columns to select when invoke `cache.fetch()`
        :param column_types:    if column not specified here, type should be str
        :param value_adapter:   convert value when fetch and set
        :param default_column_type:     if column type not specified, use this value as default
        :param encoder:         convert value when set
        :param decoder:         convert value when fetch
        """
        self.table = table
        self.columns = columns
        self.column_types = dict(column_types or {})
        self.default_column_type = default_column_type
        if isinstance(uniq_columns, str):
            self.uniq_columns = (uniq_columns,)
        else:
            self.uniq_columns = uniq_columns
        if not self.uniq_columns:
            raise ValueError("uniq_columns cannot be empty.")
        self.columns_with_id = [x for x in self.uniq_columns if x not in self.columns] + list(self.columns)
        self.value_adapter = self._check_value_adapter(value_adapter, columns)
        self.encoder = encoder
        self.decoder = decoder
        if self.encoder is None:
            if self.value_adapter is not None:
                self.encoder = self.value_adapter.build_column_values
            else:
                self.encoder = NoModifyEncoder()
        if self.decoder is None:
            if self.value_adapter is not None:
                self.decoder = self.value_adapter.parse_column_values
            else:
                self.decoder = NoModifyDecoder()

    @staticmethod
    def _check_value_adapter(adapter, columns: Sequence[str]) -> ValueAdapter:
        if adapter == 'single' and len(columns) != 1:
            raise ValueError(f"columns length should be 1 when value_adapter is 'single'.")
        if isinstance(adapter, str) and adapter in _BUILTIN_ADAPTER_BUILDERS:
            return _BUILTIN_ADAPTER_BUILDERS[adapter](columns)
        if isinstance(adapter, ValueAdapter):
            return adapter
        if isinstance(adapter, str):
            msg = f"Unexpected value_adapter: {adapter}"
        else:
            msg = f"Unexpected value_adapter type: {type(adapter)}"
        raise ValueError(msg)

    def get_column_type(self, col: str) -> str:
        return self.column_types.get(col, self.default_column_type)

    @staticmethod
    def normalize_columns_values(value: Any, columns: Sequence[str]) -> Sequence[Any]:
        n = len(columns)
        if n > 1:
            if not isinstance(value, (list, tuple)):
                raise ValueError(f"value should be a list or tuple, got {type(value)}")
            m = len(value)
            if m != n:
                raise ValueError(f"value size should be {n}, but got {m}")
            return value
        else:
            return [value]

    def normalize_uniq_column_values(self, key: Any) -> Sequence[Any]:
        return self.normalize_columns_values(key, self.uniq_columns)

class RelationalDbScopeConfig(BaseTableConfig):
    def __init__(
        self,
        scope: Optional[Hashable], table: str = None,
        uniq_column: str = None,
        uniq_columns: Sequence[str] = ('id',),
        columns: Sequence[str] = ('data',),
        column_types: Dict[str, str] = None,
        value_adapter: Union[Literal['default', 'single', 'tuple', 'list', 'json'], ValueAdapter] = 'default',
        default_column_type: str = None,
        encoder: Union[None, ValueEncoder, Callable[[Any], Dict[str, Any]]] = None,
        decoder: Union[None, ValueDecoder, Callable[[Dict[str, Any]], Any]] = None,
    ):
        self.scope = scope
        if not table:
            if not isinstance(scope, str):
                raise ValueError("Arg scope must be a str when table not given.")
            table = scope
        if uniq_column:
            warnings.warn("`uniq_column` is deprecated, use `uniq_columns` instead.", DeprecationWarning)
            uniq_columns = (uniq_column,)
        super().__init__(
            table=table,
            uniq_columns=uniq_columns,
            columns=columns,
            column_types=column_types,
            value_adapter=value_adapter,
            default_column_type=default_column_type,
            encoder=encoder,
            decoder=decoder,
        )

class SingleTableConfig(BaseTableConfig):
    def __init__(self, table: str,
                 scope_columns: Sequence[str] = ('scope',),
                 key_columns: Sequence[str] = ('id',),
                 columns: Sequence[str] = ('data',),
                 column_types: Dict[str, str] = None,
                 value_adapter: Union[Literal['auto', 'default', 'single', 'tuple', 'list'], ValueAdapter] = 'auto',
                 default_column_type: str = None):
        self.scope_columns = tuple(scope_columns)
        self.key_columns = tuple(key_columns)
        uniq_columns = self.scope_columns + self.key_columns
        super().__init__(table, uniq_columns, columns, column_types, value_adapter, default_column_type)

    def normalize_scope_column_values(self, scope: Any) -> Sequence[Any]:
        return self.normalize_columns_values(scope, self.scope_columns)

    def normalize_key_column_values(self, scope: Any) -> Sequence[Any]:
        return self.normalize_columns_values(scope, self.key_columns)

